# Copyright (C) 2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Border subcommand."""

import argparse
import logging

from PIL import Image, ImageColor, ImageOps

from fotolab import save_image

log = logging.getLogger(__name__)


def build_subparser(subparsers) -> None:
    """Build the subparser."""
    border_parser = subparsers.add_parser("border", help="add border to image")

    border_parser.set_defaults(func=run)

    border_parser.add_argument(
        dest="image_filenames",
        help="set the image filenames",
        nargs="+",
        type=str,
        default=None,
        metavar="IMAGE_FILENAMES",
    )

    border_parser.add_argument(
        "-c",
        "--color",
        dest="color",
        type=str,
        default="black",
        help="set the color of border (default: '%(default)s')",
        metavar="COLOR",
    )

    border_parser.add_argument(
        "-w",
        "--width",
        dest="width",
        type=str,
        default=10,
        help="set the width of border (default: '%(default)s')",
        metavar="WIDTH",
    )

    border_parser.add_argument(
        "-wt",
        "--width-top",
        dest="width_top",
        type=str,
        default=0,
        help="set the width of top border (default: '%(default)s')",
        metavar="WIDTH",
    )

    border_parser.add_argument(
        "-wr",
        "--width-right",
        dest="width_right",
        type=str,
        default=0,
        help="set the width of right border (default: '%(default)s')",
        metavar="WIDTH",
    )

    border_parser.add_argument(
        "-wb",
        "--width-bottom",
        dest="width_bottom",
        type=str,
        default=0,
        help="set the width of bottom border (default: '%(default)s')",
        metavar="WIDTH",
    )

    border_parser.add_argument(
        "-wl",
        "--width-left",
        dest="width_left",
        type=str,
        default=0,
        help="set the width of left border (default: '%(default)s')",
        metavar="WIDTH",
    )

    border_parser.add_argument(
        "-op",
        "--open",
        default=False,
        action="store_true",
        dest="open",
        help="open the image using default program (default: '%(default)s')",
    )

    border_parser.add_argument(
        "-od",
        "--output-dir",
        dest="output_dir",
        default="output",
        help="set default output folder (default: '%(default)s')",
    )


def run(args: argparse.Namespace) -> None:
    """Run border subcommand.

    Args:
        args (argparse.Namespace): Config from command line arguments

    Returns:
        None
    """
    log.debug(args)

    for image_filename in args.image_filenames:
        original_image = Image.open(image_filename)

        border = get_border(args)
        bordered_image = ImageOps.expand(
            original_image,
            border=border,
            fill=ImageColor.getrgb(args.color),
        )

        save_image(args, bordered_image, image_filename, "border")


def get_border(args: argparse.Namespace) -> tuple[int, int, int, int]:
    """Calculate the border dimensions.

    Args:
        args (argparse.Namespace): Command line arguments

    Returns:
        tuple[int, int, int, int]: Border dimensions in pixesl. If individual
        widths are not specified, returns a uniform width for all sides.
    """
    if any(
        [
            args.width_left,
            args.width_top,
            args.width_right,
            args.width_bottom,
        ]
    ):
        return (
            int(args.width_left),
            int(args.width_top),
            int(args.width_right),
            int(args.width_bottom),
        )
    return args.width



# Copyright (C) 2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Info subcommand."""

import argparse
import logging

from PIL import ExifTags, Image

log = logging.getLogger(__name__)


def build_subparser(subparsers) -> None:
    """Build the subparser."""
    info_parser = subparsers.add_parser("info", help="info an image")

    info_parser.set_defaults(func=run)

    info_parser.add_argument(
        dest="image_filename",
        help="set the image filename",
        type=str,
        default=None,
        metavar="IMAGE_FILENAME",
    )

    info_parser.add_argument(
        "-s",
        "--sort",
        default=False,
        action="store_true",
        dest="sort",
        help="show image info by sorted field name",
    )

    info_parser.add_argument(
        "--camera",
        default=False,
        action="store_true",
        dest="camera",
        help="show the camera maker details",
    )

    info_parser.add_argument(
        "--datetime",
        default=False,
        action="store_true",
        dest="datetime",
        help="show the datetime",
    )


def run(args: argparse.Namespace) -> None:
    """Run info subcommand.

    Args:
        args (argparse.Namespace): Config from command line arguments

    Returns:
        None
    """
    log.debug(args)

    info = []
    image = Image.open(args.image_filename)

    if args.camera:
        info.append(camera_metadata(image))

    if args.datetime:
        info.append(datetime(image))

    if info:
        print("\n".join(info))
    else:
        exif_tags = extract_exif_tags(image)
        if exif_tags:
            tag_name_width = max(map(len, exif_tags))
            for tag_name, tag_value in exif_tags.items():
                print(f"{tag_name:<{tag_name_width}}: {tag_value}")
        else:
            print("No metadata found!")


def extract_exif_tags(image: Image.Image, sort: bool = False) -> dict:
    """Extract Exif metadata from image."""
    try:
        exif = image._getexif()
    except AttributeError:
        exif = None

    log.debug(exif)

    info = {}
    if exif:
        info = {ExifTags.TAGS.get(tag_id): exif.get(tag_id) for tag_id in exif}

    filtered_info = {
        key: value for key, value in info.items() if key is not None
    }
    if sort:
        filtered_info = dict(sorted(filtered_info.items()))

    return filtered_info


def datetime(image: Image.Image):
    """Extract datetime metadata."""
    exif_tags = extract_exif_tags(image)
    return exif_tags["DateTime"]


def camera_metadata(image: Image.Image):
    """Extract camera and model metadata."""
    exif_tags = extract_exif_tags(image)
    metadata = f'{exif_tags["Make"]} {exif_tags["Model"]}'
    return metadata.strip()



# Copyright (C) 2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""A console program that manipulate images."""

import argparse
import logging
import subprocess
import sys
from pathlib import Path

from PIL import Image

__version__ = "0.28.0"

log = logging.getLogger(__name__)


def save_gif_image(
    args: argparse.Namespace,
    image_filename: str,
    original_image: Image.Image,
    after_image: Image.Image,
    subcommand: str,
) -> None:
    """Save the original and after image."""
    image_file = Path(image_filename)
    new_filename = Path(
        args.output_dir,
        image_file.with_name(f"{subcommand}_gif_{image_file.stem}.gif"),
    )
    new_filename.parent.mkdir(parents=True, exist_ok=True)

    log.info("f{subcommand} gif image: %s", new_filename)
    original_image.save(
        new_filename,
        format="gif",
        append_images=[after_image],
        save_all=True,
        duration=2500,
        loop=0,
        optimize=True,
    )

    if args.open:
        _open_image(new_filename)


def save_image(args, new_image, output_filename, subcommand):
    """Save image after image operation.

    Args:
        args (argparse.Namespace): Config from command line arguments
        new_image(PIL.Image.Image): Modified image
        output_filename(str): Save filename image
        subcommand(str): Subcommand used to call this function

    Returns:
        None
    """
    image_file = Path(output_filename)
    new_filename = _get_output_filename(args, image_file, subcommand)
    log.info("%s image: %s", subcommand, new_filename.resolve())
    new_image.save(new_filename)

    if args.open:
        _open_image(new_filename)


def _get_output_filename(args, image_file, subcommand):
    """Build and return output filename based on the command line options."""
    if args.overwrite:
        return image_file.with_name(image_file.name)

    output_dir = Path(args.output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    return output_dir / image_file.with_name(f"{subcommand}_{image_file.name}")


def _open_image(filename):
    """Open generated image using default program."""
    platform_open_func = {
        "linux": subprocess.call(["xdg-open", filename]),
        "darwin": subprocess.call(["open", filename]),
        "windows": subprocess.call(["start", filename]),
    }
    try:
        platform_open_func[sys.platform]
    except KeyError:
        print(f"Unsupported platform: {sys.platform}")
    else:
        log.info("open image: %s", filename.resolve())



# Copyright (C) 2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Halftone subcommand."""

import argparse
import logging
import math

from PIL import Image, ImageDraw

from fotolab import save_gif_image, save_image

log = logging.getLogger(__name__)


def build_subparser(subparsers) -> None:
    """Build the subparser."""
    halftone_parser = subparsers.add_parser(
        "halftone", help="halftone an image"
    )

    halftone_parser.set_defaults(func=run)

    halftone_parser.add_argument(
        dest="image_filenames",
        help="set the image filename",
        nargs="+",
        type=str,
        default=None,
        metavar="IMAGE_FILENAMES",
    )

    halftone_parser.add_argument(
        "-ba",
        "--before-after",
        default=False,
        action="store_true",
        dest="before_after",
        help="generate a GIF showing before and after changes",
    )

    halftone_parser.add_argument(
        "-op",
        "--open",
        default=False,
        action="store_true",
        dest="open",
        help="open the image using default program (default: '%(default)s')",
    )

    halftone_parser.add_argument(
        "-od",
        "--output-dir",
        dest="output_dir",
        default="output",
        help="set default output folder (default: '%(default)s')",
    )


def run(args: argparse.Namespace) -> None:
    """Run halftone subcommand.

    Args:
        args (argparse.Namespace): Config from command line arguments

    Returns:
        None
    """
    log.debug(args)

    for image_filename in args.image_filenames:
        original_image = Image.open(image_filename)
        halftone_image = create_halftone_image(original_image)

        if args.before_after:
            save_gif_image(
                args,
                image_filename,
                original_image,
                halftone_image,
                "halftone",
            )
        else:
            save_image(args, halftone_image, image_filename, "halftone")


def create_halftone_image(
    original_image: Image.Image, cell_count: int = 50
) -> Image.Image:
    """Create a halftone version of the input image.

    Modified from the circular halftone effect processing.py example from
    https://tabreturn.github.io/code/processing/python/2019/02/09/processing.py_in_ten_lessons-6.3-_halftones.html

    Args:
        original_image: The source image to convert
        cell_count: Number of cells across the width (default: 50)

    Returns:
        Image.Image: The halftone converted image
    """
    grayscale_image = original_image.convert("L")
    width, height = original_image.size

    halftone_image = Image.new("L", (width, height), "black")
    draw = ImageDraw.Draw(halftone_image)

    cellsize = width / cell_count
    rowtotal = math.ceil(height / cellsize)

    for row in range(rowtotal):
        for col in range(cell_count):
            # Calculate center point of current cell
            x = int(col * cellsize + cellsize / 2)
            y = int(row * cellsize + cellsize / 2)

            # Get brightness and calculate dot size
            brightness = grayscale_image.getpixel((x, y))
            dot_size = 10 * brightness / 200

            # Draw the dot
            draw.ellipse(
                [
                    x - dot_size / 2,
                    y - dot_size / 2,
                    x + dot_size / 2,
                    y + dot_size / 2,
                ],
                fill=255,
            )

    return halftone_image
